import paho.mqtt.client as mqtt
from threading import Event


class MQTTClient(object):
    """Wrapper for the paho.mqtt.client. Used by ADriver and DeviceManager but can also be used for other purposes."""

    client = None  # holds an instance of paho.mqtt.client.Client
    _config = None  # json configuration
    _verbose_mqttclient = False  # print debugging information if set to yes.
    on_message = None  # holds the function that the client uses upon receiving a message from a subscribed channel.
    is_connected = None  # threading.Event - True if connection to mqtt broker has been successfully established.
    is_disconnected = None

    def __init__(self, config, verbose):
        self._config = config
        self._verbose_mqttclient = verbose
        self.is_connected = Event()
        self.is_connected.clear()
        self.is_disconnected = Event()
        self.is_disconnected.set()
        self.client = mqtt.Client()

    def connect(self):
        """Connect to the mqtt broker using the provided configuration and on_message function."""
        if self._verbose_mqttclient:
            print("Connecting to mqtt.")
        self.client.on_connect = self._on_connect
        self.client.on_message = self.on_message
        self.client.username_pw_set(self._config["mqtt-user"], password=self._config["mqtt-password"])
        self.client.connect(self._config["mqtt-address"], self._config["mqtt-port"], 60)
        self.client.loop_start()

    def disconnect(self):
        """Disconnect from mqtt broker and set is_connected to False."""
        self.client.disconnect()
        self.is_connected.clear()
        self.is_disconnected.set()

    def _on_connect(self, client, userdata, flags, rc):
        """Return code after trying to connect to mqtt brokder. If successfully connected, is_connected is True."""
        if self._verbose_mqttclient:
            print("Connected with result code " + str(rc))
        if rc == 0:
            self.is_connected.set()
            self.is_disconnected.clear()

    def publish(self, topic, msg):
        if self._verbose_mqttclient:
            print("publishing to topic '{}' the message '{}'.".format(topic, msg))
        self.client.publish(topic, msg)

    @staticmethod
    def merge(driver, mqtt_client):
        """Utility method - joins a driver instance with an external mqtt-client. Must be called before connecting."""
        driver._mqtt = mqtt_client
        mqtt_client.on_message = driver._on_message


from alcathous.nodatabehavior import NoDataBehavior


class Algorithm:
    _data_set = None
    _lock_data_set = None
    _time_window = None  # time window in seconds
    _no_data_behavior = None
    _last_valid = None
    _topic = None
    _mqtt_pub_function = None
    name = None
    _config = None
    _verbose = None

    def __init__(self, name, config, verbose, data_set, lock_data_set, topic_pub_prefix, mqtt_pub_function,
                 no_data_behavior):
        self.name = name
        self._config = config
        self._verbose = verbose
        self._no_data_behavior = no_data_behavior
        self._data_set = data_set
        self._lock_data_set = lock_data_set

        if self._verbose:
            print("{}.{} - __init__".format(self.__class__.__name__, name))
        self._topic = topic_pub_prefix + self._config["topic-pub-suffix"]
        if self._verbose:
            print("{}.{} - publish to topic '{}'.".format(self.__class__.__name__, name, self._topic))

        self._mqtt_pub_function = mqtt_pub_function
        self._time_window = int(self._config["time_window"]) * 60
        if self._time_window <= 0:
            raise ValueError("Value for time_window must be a positive integer larger than 0. ('{}' not > 0)".
                             format(self._time_window))
        if self._verbose:
            print("{}.{} - time window {} s.".format(self.__class__.__name__, name, self._time_window))

    def process(self, timestamp):
        time_from = timestamp - self._time_window
        time_to = timestamp
        try:
            value = self._process(time_from, time_to)
            self._last_valid = value
            self._mqtt_pub_function(self._topic, value)
        except ValueError:
            if self._verbose:
                print("{} - process/ValueError. performing no data behavior '{}'.".
                      format(self._topic, self._no_data_behavior))
            if self._no_data_behavior == NoDataBehavior.MUTE:
                pass
            elif self._no_data_behavior == NoDataBehavior.EMPTY_MESSAGE:
                self._mqtt_pub_function(self._topic, None)
            elif self._no_data_behavior == NoDataBehavior.LAST_VALID:
                self._mqtt_pub_function(self._topic, self._last_valid)
            else:
                raise NotImplementedError("Don't know how to handle NoDataBehavior.{}.".
                                          format(self._no_data_behavior))

    def _process(self, time_from, time_to):
        raise NotImplementedError()

    def execution_points_estimation(self):
        raise NotImplementedError()


class Average(Algorithm):
    def _process(self, time_from, time_to):
        if self._verbose:
            print("{} - average for '{} s' to '{} s'.".format(self._topic, time_from, time_to))
        sum_values = 0
        count = 0
        with self._lock_data_set:
            if self._verbose:
                print("{} - acquired lock".format(self._topic))
            for timestamp, value in reversed(self._data_set.items()):
                if self._verbose:
                    print("{} - timestamp '{} s'; value '{}'.".format(self._topic, timestamp, value))
                if time_from > timestamp:
                    if self._verbose:
                        print("{} - time_from > timestamp".format(self._topic))
                    break
                elif timestamp <= time_to:
                    sum_values = sum_values + value
                    count = count + 1
        if self._verbose:
            print("{} - sum: {}; count: {}.".format(self._topic, sum_values, count))
        try:
            avg = sum_values/count
        except ZeroDivisionError:
            raise ValueError
        return avg

    def execution_points_estimation(self):
        return int(self._time_window)


class WeightedAverage(Algorithm):
    def _process(self, time_from, time_to):
        if self._verbose:
            print("{} - weighted average for '{} s' to '{} s'.".format(self._topic, time_from, time_to))
        sum_weighted_values = 0
        weighted_count = 0
        with self._lock_data_set:
            if self._verbose:
                print("{} - acquired lock".format(self._topic))
            for timestamp, value in reversed(self._data_set.items()):
                if time_from > timestamp:
                    if self._verbose:
                        print("{} - timestamp '{} s'; value '{}'.".format(self._topic, timestamp, value))
                    break
                elif timestamp <= time_to:
                    weight = timestamp - time_from
                    sum_weighted_values = sum_weighted_values + value * weight
                    weighted_count = weighted_count + weight
        if self._verbose:
            print("{} - weighted-sum: {}; weighted-count: {}.".
                  format(self._topic, sum_weighted_values, weighted_count))
        try:
            wavg = sum_weighted_values / weighted_count
        except ZeroDivisionError:
            raise ValueError
        return wavg

    def execution_points_estimation(self):
        return int(self._time_window * 1.1)


class AlgorithmFactory:
    avg = Average
    wavg = WeightedAverage

    @classmethod
    def get_instance(cls, name, config, verbose, data_set, lock_data_set, topic_pub_prefix, mqtt_pub_function,
                     no_data_behavior):
        class_name = str(config["algorithm"]).lower()
        try:
            klass = cls.__dict__[class_name]
        except KeyError:
            raise ValueError("Unknown value for AlgorithmFactory '{}'. Expected {}.".
                             format(class_name, cls.__dict__.keys()))
        if verbose:
            print("AlgorithmFactory.get_instance - creating instance of '{}'.".format(klass.__name__))
        instance = klass(name, config, verbose, data_set, lock_data_set, topic_pub_prefix, mqtt_pub_function,
                         no_data_behavior)
        return instance

    @classmethod
    def get_instances(cls, method_names, config_methods, verbose, data_set, lock_data_set, topic_pub_prefix,
                      mqtt_pub_function, no_data_behavior):
        if verbose:
            print("AlgorithmFactory.get_instances - creating instances of '{}'.".format(method_names))
        methods = []
        for name in method_names:
            try:
                config = config_methods[name]
            except KeyError:
                raise ValueError("Unknown value for method '{}'. Expected {}.".
                                 format(name, config_methods.keys()))
            m = AlgorithmFactory.get_instance(name, config, verbose, data_set, lock_data_set,
                                              topic_pub_prefix, mqtt_pub_function, no_data_behavior)
            methods.append(m)
        return methods



# this module removes yaml1.1 compliant exchange of On/Off & Yes/No with True/False.
# as soon as pyyaml supports yaml1.2 this is deprecated

from yaml import load  # to be used by adriver.py and devicemanager.py
from yaml.resolver import Resolver
try:
    from yaml import CLoader as Loader, CDumper as Dumper
except ImportError:
    from yaml import Loader, Dumper

# remove resolver entries for On/Off/Yes/No
# (https://stackoverflow.com/questions/36463531/pyyaml-automatically-converting-certain-keys-to-boolean-values)
for ch in "OoYyNn":
    if len(Resolver.yaml_implicit_resolvers[ch]) == 1:
        del Resolver.yaml_implicit_resolvers[ch]
    else:
        Resolver.yaml_implicit_resolvers[ch] = [x for x in
                Resolver.yaml_implicit_resolvers[ch] if x[0] != 'tag:yaml.org,2002:bool']


def dict_deepcopy_lowercase(dict_in):
    if type(dict_in) is dict:
        dict_out = {}
        for k, v in dict_in.items():
            try:
                k = k.lower()
            except AttributeError:
                pass
            dict_out[k] = dict_deepcopy_lowercase(v)
    elif type(dict_in) is list:
        dict_out = []
        for d in dict_in:
            dict_out.append(dict_deepcopy_lowercase(d))
    else:
        dict_out = dict_in
    return dict_out



import argparse
import os
from pathlib import Path
import time
import threading
import queue
import os
import sys
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))
from alcathous.datapoint import DataPoint
import alcathous.mqttclient
from alcathous.nodatabehavior import NoDataBehavior
import alcathous.mypyyaml


class DataPointManager:
    _data_points = None
    _processes = None
    _purges = None
    _mqtt_client = None
    _config = None
    _verbose = None
    _mqtt_sub_handler = None
    _no_data_behavior = None
    _update_cycle = None
    _stop_loop = None
    _worker_list = None
    _worker_queue = None
    _worker_number = None
    _is_stopped = None

    def __init__(self, config, verbose):
        self._config = config
        self._verbose = verbose

        if self._verbose:
            print(self._config)

        self._purges = []
        self._processes = []
        self._data_points = []

        self._stop_loop = threading.Event()
        self._stop_loop.clear()
        self._is_stopped = threading.Event()
        self._is_stopped.set()

        self._update_cycle = int(self._config["general"]["update_cycle"])

        self._no_data_behavior = NoDataBehavior.get_enum(str(self._config["general"]["no_data_behavior"]))

        self._mqtt_client = alcathous.mqttclient.MQTTClient(self._config["mqtt"], self._verbose)
        self._mqtt_client.on_message = self._on_message
        self._mqtt_sub_handler = {}

        for config_data_point in self._config["datapoints"]:
            dp = DataPoint(config_data_point, self._config["methods"], self._verbose, self._mqtt_client.publish,
                           self._mqtt_sub_handler, self._no_data_behavior)
            self._purges.append(dp.purge_old_values)
            for method in dp.methods:
                process = method.process
                cost = method.execution_points_estimation()
                if self._verbose:
                    print("DataPointManager - adding process '{}' with cost '{}'.".format(process.__name__, cost))
                self._processes.append((process, cost))
            self._data_points.append(dp)

        self._processes.sort(key=lambda tup: tup[1], reverse=True)  # sort processes by their cost most expensive first

        self._worker_queue = queue.Queue()
        self._worker_list = []
        self._worker_number = int(self._config["general"]["number_worker"])

    def _on_message(self, client, userdata, msg):
        """Method for mqtt client on_message."""
        if self._verbose:
            print("DataPointManager - received msg at topic '{}'.".format(msg.topic))
        self._mqtt_sub_handler[msg.topic](msg.payload)

    def _worker(self):
        if self._verbose:
            print("DataPointManager - started worker")
        while True:
            item = self._worker_queue.get()
            if self._verbose:
                print("DataPointManager - worker received item '{}'.".format(item))
            if item is None:
                break
            func, parameter = item
            func(parameter)
            self._worker_queue.task_done()
        if self._verbose:
            print("DataPointManager - stopped worker")

    def _do_loop(self):
        timestamp = time.time()
        if self._verbose:
            print("DataPointManager - started work for timestamp '{} s'.".format(timestamp))
        for p in self._processes:
            self._worker_queue.put((p[0], timestamp))
        if self._verbose:
            print("DataPointManager - waiting for worker to finish processing the algorithms.")
        self._worker_queue.join()
        for p in self._purges:
            self._worker_queue.put((p, timestamp))
        if self._verbose:
            print("DataPointManager - waiting for worker to purge outdated values.")
        self._worker_queue.join()

    def _start(self):
        if self._verbose:
            print("DataPointManager - start loop.")
        self._is_stopped.clear()
        for i in range(self._worker_number):
            w = threading.Thread(target=self._worker)
            w.start()
            self._worker_list.append(w)
        self._mqtt_client.connect()
        self._mqtt_client.is_connected.wait()

        for topic in self._mqtt_sub_handler.keys():
            print("DataPointManager - subscribing to topic '{}'.".format(topic))
            self._mqtt_client.client.subscribe(topic)

        while not self._stop_loop.is_set():
            start = time.time()
            self._do_loop()
            sleep_for = max(0, self._update_cycle - (time.time() - start))
            if self._verbose:
                print("DataPointManager - wait for '{} s'.".format(sleep_for))
            self._stop_loop.wait(sleep_for)
        if self._verbose:
            print("DataPointManager - exited loop.")

    def _stop(self):
        if self._verbose:
            print("DataPointManager - stop loop.")
        self._stop_loop.set()
        self._mqtt_client.disconnect()
        self._mqtt_client.is_disconnected.wait()
        if self._verbose:
            print("DataPointManager - stopping worker.")
        for i in len(self._worker_list):
            self._worker_queue.put(None)
        for t in self._worker_list:
            t.join()
        self._worker_list = []
        self._is_stopped.set()
        if self._verbose:
            print("DataPointManager - loop stopped.")

    @classmethod
    def _args_to_config(cls, args=None):
        """Handle command line arguments and read the yaml file into a json structure (=config)."""
        desc = "DataPoint Manager"
        ap = argparse.ArgumentParser(description=desc)
        ap.add_argument('-c', '--config', type=str, help='yaml config file', required=True)
        ap.add_argument('-v', '--verbose', help='verbose', action="store_true")
        ap.add_argument('--version', action='version',
                            version='%(prog)s {}'.format(alcathous.version),
                            help='show the version number and exit')
        if args:
            arguments = vars(ap.parse_args(args))
        else:
            arguments = vars(ap.parse_args())

        verbose = False
        if arguments["verbose"]:
            verbose = True

        config_filename = arguments["config"]
        config_file = Path(config_filename)
        if not config_file.is_file():
            raise FileNotFoundError("config file '{}' not found.".format(config_filename))

        config = alcathous.mypyyaml.load(open(config_filename, 'r'), Loader=alcathous.mypyyaml.Loader)
        credentials = alcathous.mypyyaml.load(open(os.path.expanduser(config["mqtt"]["mqtt-credentials"]), 'r'),
                                    Loader=alcathous.mypyyaml.Loader)
        config["mqtt"].update(credentials)

        return config, verbose

    @classmethod
    def standalone(cls, args=None):
        config, verbose = DataPointManager._args_to_config(args)
        config = alcathous.mypyyaml.dict_deepcopy_lowercase(config)
        dm = DataPointManager(config, verbose)
        dm._start()
        try:
            while not dm._is_stopped.wait(0.1):  # timeout is necessary for CTRL+C
                pass
        except KeyboardInterrupt:
            pass
        dm._stop()


def standalone():
    DataPointManager.standalone()


if __name__ == "__main__":
    DataPointManager.standalone()


import collections
from threading import Lock
import time
from alcathous.algorithms import AlgorithmFactory


class DataPoint:
    methods = None
    _topic_sub = None
    _topic_pub_prefix = None
    _zero_is_valid = None
    _max_time_window = None
    _data_set = None
    _lock_data_set = None
    _verbose = None

    def __init__(self, config_datapoint, config_methods, verbose, mqtt_pub_function, mqtt_sub_handler,
                 no_data_behavior):
        self._verbose = verbose
        self._topic_sub = str(config_datapoint["topic-sub"])
        if self._verbose:
            print("{} - __init__".format(self._topic_sub))
        self._topic_pub_prefix = str(config_datapoint["topic-pub-prefix"])
        if self._verbose:
            print("{} - publish to '{}#'".format(self._topic_sub, self._topic_pub_prefix))
        self._data_set = collections.OrderedDict()
        self._lock_data_set = Lock()
        self._max_time_window = 0
        self._zero_is_valid = bool(config_datapoint["zero_is_valid"])
        mqtt_sub_handler[self._topic_sub] = self._message_handler

        temp_methods = [x.strip() for x in config_datapoint["methods"].split(',')]
        self.methods = AlgorithmFactory.get_instances(temp_methods, config_methods, self._verbose, self._data_set,
                                                      self._lock_data_set, self._topic_pub_prefix, mqtt_pub_function,
                                                      no_data_behavior)
        for m in self.methods:
            if m._time_window > self._max_time_window:
                self._max_time_window = m._time_window

        if self._verbose:
            print("{} - max time window for purging data: {} s.".format(self._topic_sub, self._max_time_window))

    def _message_handler(self, value):
        if self._is_value_valid(value):
            with self._lock_data_set:
                timestamp = time.time()
                self._data_set[timestamp] = float(value)
                if self._verbose:
                    print("{} - added {}@{}s".format(self._topic_sub, value, timestamp))
                    print(self._data_set)

    def _is_value_valid(self, value):
        result = True
        if value is None:
            result = False
        elif value == 0 and not self._zero_is_valid:
            result = False

        if not result and self._verbose:
            print("{} - value '{}' is not valid.".format(self._topic_sub, value))
        return result

    def purge_old_values(self, timestamp):
        min_time_stamp = timestamp - self._max_time_window
        if self._verbose:
            print("{} - purging values with timestamp < '{}'.".format(self._topic_sub, min_time_stamp))
            count = 0
            list_size = len(self._data_set)
        with self._lock_data_set:
            while len(self._data_set) and (next(iter(self._data_set.items())))[0] < min_time_stamp:
                if self._verbose:
                    print("{} - purge item '{}'.".format(self._topic_sub, (next(iter(self._data_set.items())))))
                    count = count + 1
                self._data_set.popitem(False)
        if self._verbose:
            print("{} -  purged {}/{} items.".format(self._topic_sub, count, list_size))





from enum import Enum


class NoDataBehavior(Enum):
    MUTE = 0,
    LAST_VALID = 1,
    EMPTY_MESSAGE = 2

    @classmethod
    def get_enum(cls, config_str):
        config_str = str(config_str).upper()
        try:
            result = cls.__members__[config_str]
        except KeyError:
            raise ValueError("Unknown value for NoDataBehavior '{}'. Expected {}.".format(config_str, cls.__members__.keys()))
        return result
